/*
 * File:   main.c
 * Author: krikzz
 *
 * Created on 23.06.2010
 */



#include "../../lib/segalib.h"

#define MAX_TESTS (13)

static u16 hvc[4];
static u16 joy;
static u8 test;
static u8 sel;
static u8 hMode;

static void val2str(u16 hvc, char *str)
{
    u8 val;
	  val = (hvc >> 12) & 0x0F;
    if (val < 10) str[1] = 0x30 + val;
    else str[1] = 0x41 + val - 10;
    val = (hvc >> 8) & 0x0F;
    if (val < 10) str[2] = 0x30 + val;
    else str[2] = 0x41 + val - 10;
    val = (hvc >> 4) & 0x0F;
    if (val < 10) str[3] = 0x30 + val;
    else str[3] = 0x41 + val - 10;
    val = (hvc >> 0) & 0x0F;
    if (val < 10) str[4] = 0x30 + val;
    else str[4] = 0x41 + val - 10;
	  str[0]='$';
	  str[5]=0;
}

static void test_1(void)
{
  char msg[10];
	
	msg[0] = 'H';
	msg[1] = 'V';
	msg[2] = 'C';
	msg[3] = '=';

	hvc[0] = 0;
	hvc[1] = 0;
	hvc[2] = 0;
	hvc[3] = 0;

	// enable HINT
	test = 0;
	VDP_setReg(10,0x80);
	VDP_setHInterrupt(1);

	// wait for HINT counter to be reloaded
	VDP_waitVSync();
	
	// enable HINT callback
	test = 1;

	//wait for VBLANK
	VDP_waitVSync(); 
	
  hvc[2] = GET_HVCOUNTER; // read HVC
	VDP_setReg(0,0x04);     // disable HVC latch
  hvc[3] = GET_HVCOUNTER; // read HVC
	
  //clean screen
	VDP_fillTileRect(APLAN, 0, 0, 0, 40, 27);
        
  //display HVC read values
	val2str(hvc[0],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 12);
	val2str(hvc[1],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 13);
	val2str(hvc[2],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 14);
	val2str(hvc[3],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 15);

	VDP_drawText(APLAN, "Press START to quit", 0, 2,26);
	while (!(joy & BUTTON_START)) VDP_waitVSync();
}

static void test_2(u8 vint)
{
  u16 status = 0;
  char msg[10];
	
	msg[0] = 'H';
	msg[1] = 'V';
	msg[2] = 'C';
	msg[3] = '=';

  // clear HINT pending flag
  VDP_setReg(10,0xFF);
  VDP_setReg(0,0x14);
	VDP_waitVSync(); 
	VDP_waitVSync(); 

  // wait VINT pending flag to be cleared
  while (!status) status = GETVDPSTATUS(VDP_VINTPENDING_FLAG);
  while (status) status = GETVDPSTATUS(VDP_VINTPENDING_FLAG);

  // mask all 68k interrupts
  asm("move	#0x2700,%sr");

	// enable HVC latch + HINT or VINT
  if (vint)
  {
	  VDP_setReg(0,0x06);
    VDP_setReg(1,0x64);
	}
  else
  {
	  VDP_setReg(0,0x16);
    VDP_setReg(1,0x44);
	  VDP_setReg(10,0xC0);
	}

	// read HVC value as reference
  hvc[0] = GET_HVCOUNTER;

  // set TH as input on PORT 2 and disable /HL output (for test 2)
  volatile u8 *pb = (volatile u8 *) (0xA1000B);
  *pb = 0x80;

	// wait for interrupt to occur
  while (!status) status = GETVDPSTATUS(VDP_VINTPENDING_FLAG);

	// read new HVC value
  hvc[1] = GET_HVCOUNTER;
        
  //clean screen
	VDP_fillTileRect(APLAN, 0, 0, 0, 40, 27);

  // display infos
	val2str(hvc[0],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 12);
	val2str(hvc[1],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 13);

  // unmask interrupts
  asm("move	#0x2300,%sr");

  // enable VINT & disable HVC latch
  VDP_setReg(0,0x04);
  VDP_setReg(1,0x64);

  VDP_drawText(APLAN, "Press START to quit", 0, 2, 26);
	while (!(joy & BUTTON_START)) VDP_waitVSync();
}

static void test_3(u8 interrupt_line, u8 cpu)
{
  char msg[10];
	u8 line;

	msg[0] = 'H';
	msg[1] = 'V';
	msg[2] = 'C';
	msg[3] = '=';

	// enable HINT
	VDP_setReg(10,0x80);
	VDP_setHInterrupt(1);

	// wait for HINT to be cleared
	VDP_waitVSync();

  // disable HINT
	if (cpu) asm("move	#0x2500,%sr");
  else VDP_setHInterrupt(0);
	hvc[0] = 0;
	hvc[1] = 0;
	hvc[2] = 0;
	hvc[3] = 0;

	//wait for line 0
	line = GET_VCOUNTER;
  while (line != 0x00) line = GET_VCOUNTER;

  // disable HINT completely
	VDP_setReg(10,0xFF);

  //wait for interrupt line + 1
	line = GET_VCOUNTER;
  while (line != 0xC1) line = GET_VCOUNTER;

  // save HVC value (should remain zero)
  hvc[1] = hvc[0];


	//wait for given line
  while (line != interrupt_line) line = GET_VCOUNTER;

  // disable HINT (HINT pending flag should remain set)
	VDP_setHInterrupt(0);

	// enable HINT
  VDP_setHInterrupt(1);
	if (cpu) asm("move	#0x2300,%sr");

	//wait a few lines for HINT to occurs
  while (line !=  (interrupt_line + 5)) line = GET_VCOUNTER;

  // save HVC value
  hvc[2] = hvc[0];

	// wait for BLANK
	VDP_waitVSync();
	
  //clean screen
	VDP_fillTileRect(APLAN, 0, 0, 0, 40, 27);
        
  //display HVC read values
	val2str(hvc[1],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 12);
	val2str(hvc[2],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 13);

	VDP_drawText(APLAN, "Press START to quit", 0, 2,26);
	while (!(joy & BUTTON_START)) VDP_waitVSync();
}

static void test_4(u8 interrupt_line, u8 cpu)
{
  char msg[10];
	u8 line;
  u16 status[4];

	msg[0] = 'H';
	msg[1] = 'V';
	msg[2] = 'C';
	msg[3] = '=';

  status[0] = 0;
  status[1] = 0;
  status[2] = 0;
  status[3] = 0;

	// disable VINT
	if (cpu) asm("move	#0x2700,%sr");
  else VDP_setReg(1,0x44);
	hvc[1] = 0;
	hvc[2] = 0;
	hvc[3] = 0;

  // enable HVC read in VINT callback
  test = 0;

  //wait for interrupt line + 1
	line = GET_VCOUNTER;
  while (line != 0xE1) line = GET_VCOUNTER;

  // save HVC value (should remain zero)
  hvc[2] = hvc[1];

  // read VDP status
  status[0] = GETVDPSTATUS(0x3ff);

  //wait for given line
  while (line !=  interrupt_line) line = GET_VCOUNTER;

  // read VDP status
  status[1] = GETVDPSTATUS(0x3ff);

  // disable VINT (VINT pending flag should remain set)
	VDP_setReg(1,0x44);

  // read VDP status
  status[2] = GETVDPSTATUS(0x3ff);

	// enable VINT
	VDP_setReg(1,0x64);
  if (cpu) asm("move	#0x2300,%sr");

	//wait a few lines for VINT to occurs
  while (line != (interrupt_line + 5)) line = GET_VCOUNTER;

  // save HVC value
  hvc[3] = hvc[1];

  // read VDP status
  status[3] = GETVDPSTATUS(0x3ff);

	// wait for BLANK
	VDP_waitVSync();
	
  //clean screen
	VDP_fillTileRect(APLAN, 0, 0, 0, 40, 27);
        
  //display HVC read values
	val2str(hvc[2],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 12);
	val2str(hvc[3],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 13);

  // display VDP status
  msg[0] = 'V';
	msg[1] = 'D';
	msg[2] = 'P';
  val2str(status[0],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 15);
  val2str(status[1],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 16);
  val2str(status[2],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 17);
  val2str(status[3],msg+4);
	VDP_drawText(APLAN, msg, 0, 8, 18);

	VDP_drawText(APLAN, "Press START to quit", 0, 2,26);
	while (!(joy & BUTTON_START)) VDP_waitVSync();
}

static u8 main_menu(void)
{
    u8 ret = 0;
    u16 new_joy = 0;
    u16 old_joy = 0;

    // set TH as input on PORT 2 and disable /HL output (for test 2)
    volatile u8 *pb = (volatile u8 *) (0xA1000B);
    *pb = 0;

    //wait for VBLANK
    VDP_waitVSync();

    //clean left column
    VDP_fillTileRect(APLAN, 0, 0, 0, 40, 27);

    // display infos
		VDP_drawText(APLAN, "VDP interrupts & hvc tests", 0, 2, 8);
		VDP_drawText(APLAN, " 1: HVC latch (M3 bit)", 0, 2, 10);
    VDP_drawText(APLAN, " 2: HINT hvc (hardware mod)", 0, 2, 11);
    VDP_drawText(APLAN, " 3: VINT hvc (hardware mod)", 0, 2, 12);
		VDP_drawText(APLAN, " 4: HINT pending (current)", 0, 2, 13);
    VDP_drawText(APLAN, " 5: HINT pending (vblank)", 0, 2, 14);
		VDP_drawText(APLAN, " 6: HINT pending (next frame)", 0, 2, 15);
    VDP_drawText(APLAN, " 7: VINT pending (vblank)", 0, 2, 16);
		VDP_drawText(APLAN, " 8: VINT pending (next frame)", 0, 2, 17);
		VDP_drawText(APLAN, " 9: IRQ4 pending (current)", 0, 2, 18);
    VDP_drawText(APLAN, "10: IRQ4 pending (vblank)", 0, 2, 19);
		VDP_drawText(APLAN, "11: IRQ4 pending (next frame)", 0, 2, 20);
    VDP_drawText(APLAN, "12: IRQ6 pending (vblank)", 0, 2, 21);
		VDP_drawText(APLAN, "13: IRQ6 pending (next frame)", 0, 2, 22);
    VDP_drawText(APLAN, "Press START to run test", 0, 2, 26);

    while (!ret)
    {
        //wait for VBLANK
        VDP_waitVSync();

        //clean left column
        VDP_fillTileRect(APLAN, 1, 0, 0, 2, 27);
            
        // mark selected item
        VDP_drawText(APLAN, ">", 0, 1, 10 + sel);

        // Hmode
        if (hMode) 
        {
          VDP_setWidth256();
          VDP_drawText(APLAN, "Press A to swap mode (H32)", 0, 2, 25);
        }
        else
        {
          VDP_setWidth320();
          VDP_drawText(APLAN, "Press A to swap mode (H40)", 0, 2, 25);
        }

        // get JOY triggered events
        new_joy = (joy ^ old_joy) & joy;
        old_joy = joy;

        // change selected item
        if ((new_joy & BUTTON_UP) && (sel > 0)) sel --;
        else if ((new_joy & BUTTON_DOWN) && (sel < MAX_TESTS - 1)) sel ++;
        
        // swap mode
        if (new_joy & BUTTON_A) hMode ^= 1;

        // select test
        if (joy & BUTTON_START) ret = sel + 1;
    }
	
    // wait for START button to be released
    while (joy & BUTTON_START) VDP_waitVSync();

    //clean screen
    VDP_fillTileRect(APLAN, 0, 0, 0, 40, 27);

    return ret;
}

int main()
{
 	VDP_init();
	JOY_init();
  VDP_loadFont(font_base, 0);
  VDP_setPaletteColor(0, 15, 0xfff);//font color
  VDP_setPaletteColor(0,  0, 0x505);//border color

  sel = 0;
  hMode=0;

  while (1)
	{ 
    //main menu
		test = main_menu();
	
    switch (test)
    {
      case 1:
        test_1();
        break;
			
      case 2:
      case 3:
        test_2(test & 1);
        break;
			
      case 4:
      case 9:
        test_3(0xD0,(test > 4));
        break;
			
      case 5:
      case 10:
        test_3(0xE2,(test > 5));
        break;
			
      case 6:
      case 11:
        test_3(0x10,(test > 6));
        break;
			
      case 7:
      case 12:
        test_4(0xE4,(test > 7));
        break;
			
      case 8:
      case 13:
        test_4(0x10,(test > 8));
        break;

      default:
        break;
    }

		// wait for START button to be released
    while (joy & BUTTON_START) VDP_waitVSync();
  }
  return 0;
}

void vb()
{
  u16 val = GET_HVCOUNTER;   // read HVC

  //vblank interrupt
	joy = JOY_readJoypad(JOY_1);
  if (test == 0)
	{
    hvc[1] = val; // save HVC
  }
}

void hb()
{
  //hblank interrupt
	hvc[0] = GET_HVCOUNTER;   // read HVC
  if (test == 1)
	{
    VDP_setReg(0,0x06);     // enable HVC latch
    hvc[1] = GET_HVCOUNTER; // read HVC
  }
}

void in()
{
  //all other interrupts
}
